def _check_for_prefect_extras():
    try:
        import prefect  # noqa F401
    except ImportError:
        install_command = "pip install rubicon[prefect]"
        message = f"Install `prefect` with `{install_command}`."

        raise ImportError(message)


_check_for_prefect_extras()

from rubicon.workflow.prefect.tasks import (  # noqa E402
    create_experiment_task,
    get_or_create_project_task,
    log_artifact_task,
    log_dataframe_task,
    log_feature_task,
    log_metric_task,
    log_parameter_task,
)

__all__ = [
    "create_experiment_task",
    "get_or_create_project_task",
    "log_artifact_task",
    "log_dataframe_task",
    "log_feature_task",
    "log_metric_task",
    "log_parameter_task",
]



from dataclasses import dataclass, field
from datetime import datetime

from rubicon.domain.utils import uuid

DIRECTIONALITY_VALUES = ["score", "loss"]


@dataclass(frozen=True)
class Metric:
    name: str
    value: float

    id: str = field(default_factory=uuid.uuid4)
    description: str = None
    directionality: str = "score"
    created_at: datetime = field(default_factory=datetime.utcnow)

    def __post_init__(self):
        if self.directionality not in DIRECTIONALITY_VALUES:
            raise ValueError(f"metric directionality must be one of {DIRECTIONALITY_VALUES}")



from rubicon.client import Rubicon as SyncRubicon
from rubicon.client.asynchronous import Config, Project
from rubicon.exceptions import RubiconException


class Rubicon(SyncRubicon):
    """The asynchronous `rubicon` client's entry point.

    Creates a `Config` and injects it into the client level
    objects at run-time.

    Parameters
    ----------
    persistence : str, optional
        The persistence type. Can be one of ["filesystem"].
    root_dir : str, optional
        Absolute or relative filepath. Currently, only s3
        paths are supported asynchronously.
    auto_git_enabled : bool, optional
        True to use the `git` command to automatically log relevant repository
        information to projects and experiments logged with this client instance,
        False otherwise. Defaults to False.
    """

    def __init__(self, persistence="filesystem", root_dir=None, auto_git_enabled=False):
        self.config = Config(persistence, root_dir, auto_git_enabled)

    async def create_project(self, name, description=None, github_url=None, training_metadata=None):
        """Overrides `rubicon.client.Rubicon.create_experiment`
        to asynchronously create a project.

        Parameters
        ----------
        name : str
            The project's name.
        description : str, optional
            The project's description.
        github_url : str, optional
            The URL of the GitHub repository associated with this
            project. If omitted and automatic `git` logging is
            enabled, it will be retrieved via `git remote`.
        training_metadata : tuple or list of tuples, optional
            Metadata associated with the training dataset(s)
            used across each experiment in this project.

        Returns
        -------
        rubicon.client.asynchronous.Project
            The created project.
        """
        project = self._create_project_domain(name, description, github_url, training_metadata)
        await self.repository.create_project(project)

        return Project(project, self.config)

    async def get_project(self, name):
        """Overrides `rubicon.client.Rubicon.get_project`
        to asynchronously get a project.

        Parameters
        ----------
        name : str
            The name of the project to get.

        Returns
        -------
        rubicon.client.asynchronous.Project
            The project with name `name`.
        """
        project = await self.repository.get_project(name)

        return Project(project, self.config)

    async def get_project_as_dask_df(self, name, group_by=None):
        """Overrides `rubicon.client.Rubicon.get_project_as_dask_df`
        to asynchronously get a dask dataframe representation of a project.

        Parameters
        ----------
        name : str
            The name of the project to get.
        group_by : str or None, optional
            How to group the project's experiments in the returned
            DataFrame(s). Valid options include ["commit_hash"].

        Returns
        -------
        dask.DataFrame or list of dask.DataFrame
            If `group_by` is `None`, a dask dataframe holding the project's
            data. Otherwise a list of dask dataframes holding the project's
            data grouped by `group_by`.
        """
        project = await self.get_project(name)

        return await project.to_dask_df(group_by=group_by)

    async def get_or_create_project(self, name, **kwargs):
        """Overrides `rubicon.client.Rubicon.get_or_create_project`
        to asynchronously get or create a project.

        Parameters
        ----------
        name : str
            The project's name.
        kwargs : dict
            Additional keyword arguments to be passed to
            `Rubicon.create_project`.

        Returns
        -------
        rubicon.client.asynchronous.Project
            The corresponding project.
        """
        try:
            project = await self.get_project(name)
        except RubiconException:
            project = await self.create_project(name, **kwargs)

        return project

    async def projects(self):
        """Overrides `rubicon.client.Rubicon.projects` to
        asynchronously get a list of available projects.

        Returns
        -------
        list of rubicon.client.Project
            The list of available projects.
        """
        return [Project(project, self.config) for project in await self.repository.get_projects()]



from rubicon.client import Base, TagMixin


class Dataframe(Base, TagMixin):
    """A client dataframe.

    A `dataframe` is a two-dimensional, tabular dataset with
    labeled axes (rows and columns) that provides value to the
    model developer and/or reviewer when visualized.

    For example, confusion matrices, feature importance tables
    and marginal residuals can all be logged as a `dataframe`.

    A `dataframe` is logged to a `project` or an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Dataframe
        The dataframe domain model.
    parent : rubicon.client.Project or rubicon.client.Experiment
        The project or experiment that the artifact is
        logged to.
    """

    def __init__(self, domain, parent):
        super().__init__(domain, parent._config)

        self._data = None
        self._parent = parent

    def _get_data(self):
        """Loads the data associated with this Dataframe
        into a `dask.dataframe.DataFrame`.
        """
        project_name, experiment_id = self.parent._get_parent_identifiers()

        self._data = self.repository.get_dataframe_data(
            project_name, self.id, experiment_id=experiment_id
        )

    def plot(self, kind="table", **kwargs):
        """Render the dataframe using `mlviews`.
        """
        raise NotImplementedError

    @property
    def id(self):
        """Get the dataframe's id.
        """
        return self._domain.id

    @property
    def description(self):
        """Get the dataframe's description.
        """
        return self._domain.description

    @property
    def created_at(self):
        """Get the time this dataframe was created.
        """
        return self._domain.created_at

    @property
    def data(self):
        """Get the dataframe's raw data loaded into a
        `dask.dataframe.DataFrame`.
        """
        if self._data is None:
            self._get_data()

        return self._data

    @property
    def parent(self):
        """Get the dataframe's parent client object.
        """
        return self._parent



import os

import fsspec
from dask import dataframe as dd

from rubicon import domain
from rubicon.exceptions import RubiconException
from rubicon.repository.utils import json, slugify


class BaseRepository:
    """The base repository defines all the shared interactions
    between the various Rubicon persistence options.

    `BaseRepositoy` itself should never be used directly. Use
    one of the repositories that extends this class to persist
    Rubicon data:

        * `rubicon.repository.MemoryRepository`
        * `rubicon.repository.LocalRepository`
        * `rubicon.repository.S3Repository`

    Parameters
    ----------
    root_dir : str
        Absolute path to the root directory to persist Rubicon
        data to.
    """

    def __init__(self, root_dir):
        self.filesystem = fsspec.filesystem(self.PROTOCOL)
        self.root_dir = root_dir.rstrip("/")

    def _ls_directories_only(self, path):
        """Returns the names of all the directories at path `path`.
        """
        return [
            os.path.join(p.get("name"), "metadata.json")
            for p in self.filesystem.ls(path, detail=True)
            if p.get("type", p.get("StorageClass")).lower() == "directory"
        ]

    # -------- Projects --------

    def _get_project_metadata_path(self, project_name):
        """Returns the path of the project with name `project_name`'s
        metadata.
        """
        return f"{self.root_dir}/{slugify(project_name)}/metadata.json"

    def create_project(self, project):
        """Persist a project to the configured filesystem.

        Parameters
        ----------
        project : rubicon.domain.Project
            The project to persist.
        """
        project_metadata_path = self._get_project_metadata_path(project.name)

        if self.filesystem.exists(project_metadata_path):
            raise RubiconException(f"A project with name '{project.name}' already exists.")

        self._persist_domain(project, project_metadata_path)

    def get_project(self, project_name):
        """Retrieve a project from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project to retrieve.

        Returns
        -------
        rubicon.domain.Project
            The project with name `project_name`.
        """
        project_metadata_path = self._get_project_metadata_path(project_name)

        try:
            project = json.loads(self.filesystem.cat(project_metadata_path))
        except FileNotFoundError:
            raise RubiconException(f"No project with name '{project_name}' found.")

        return domain.Project(**project)

    def get_projects(self):
        """Get the list of projects from the filesystem.

        Returns
        -------
        list of rubicon.domain.Project
            The list of projects from the filesystem.
        """
        try:
            project_metadata_paths = self._ls_directories_only(self.root_dir)
            projects = [
                domain.Project(**json.loads(metadata))
                for metadata in self.filesystem.cat(project_metadata_paths).values()
            ]
        except FileNotFoundError:
            return []

        return projects

    # ------ Experiments -------

    def _get_experiment_metadata_root(self, project_name):
        """Returns the experiments directory of the project with
        name `project_name`.
        """
        return f"{self.root_dir}/{slugify(project_name)}/experiments"

    def _get_experiment_metadata_path(self, project_name, experiment_id):
        """Returns the path of the experiment with ID
        `experiment_id`'s metadata.
        """
        experiment_metadata_root = self._get_experiment_metadata_root(project_name)

        return f"{experiment_metadata_root}/{experiment_id}/metadata.json"

    def create_experiment(self, experiment):
        """Persist an experiment to the configured filesystem.

        Parameters
        ----------
        experiment : rubicon.domain.Experiment
            The experiment to persist.
        """
        experiment_metadata_path = self._get_experiment_metadata_path(
            experiment.project_name, experiment.id
        )

        self._persist_domain(experiment, experiment_metadata_path)

    def get_experiment(self, project_name, experiment_id):
        """Retrieve an experiment from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment to retrieve.

        Returns
        -------
        rubicon.domain.Experiment
            The experiment with ID `experiment_id`.
        """
        experiment_metadata_path = self._get_experiment_metadata_path(project_name, experiment_id)

        try:
            open_file = self.filesystem.open(experiment_metadata_path)
        except FileNotFoundError:
            raise RubiconException(f"No experiment with id `{experiment_id}` found.")

        with open_file as f:
            experiment = json.load(f)

        return domain.Experiment(**experiment)

    def get_experiments(self, project_name):
        """Retrieve all experiments from the configured filesystem
        that belong to the project with name `project_name`.

        Parameters
        ----------
        project_name : str
            The name of the project to retrieve all experiments
            from.

        Returns
        -------
        list of rubicon.domain.Experiment
            The experiments logged to the project with name
            `project_name`.
        """
        experiment_metadata_root = self._get_experiment_metadata_root(project_name)

        try:
            experiment_metadata_paths = self._ls_directories_only(experiment_metadata_root)
            experiments = [
                domain.Experiment(**json.loads(data))
                for data in self.filesystem.cat(experiment_metadata_paths).values()
            ]
        except FileNotFoundError:
            return []

        return experiments

    # ------- Artifacts --------

    def _get_artifact_metadata_root(self, project_name, experiment_id=None):
        """Returns the artifacts directory of the project with name
        `project_name` or experiment with ID `experiment_id`.
        """
        if experiment_id is not None:
            experiment_metadata_root = self._get_experiment_metadata_root(project_name)

            return f"{experiment_metadata_root}/{experiment_id}/artifacts"
        else:
            return f"{self.root_dir}/{slugify(project_name)}/artifacts"

    def _get_artifact_metadata_path(self, project_name, experiment_id, artifact_id):
        """Returns the path of the artifact with ID `artifact_id`'s
        metadata.
        """
        artifact_metadata_root = self._get_artifact_metadata_root(project_name, experiment_id)

        return f"{artifact_metadata_root}/{artifact_id}/metadata.json"

    def _get_artifact_data_path(self, project_name, experiment_id, artifact_id):
        """Returns the path of the artifact with ID `artifact_id`'s
        raw data.
        """
        artifact_metadata_root = self._get_artifact_metadata_root(project_name, experiment_id)

        return f"{artifact_metadata_root}/{artifact_id}/data"

    def create_artifact(self, artifact, data, project_name, experiment_id=None):
        """Persist an artifact to the configured filesystem.

        Parameters
        ----------
        artifact : rubicon.domain.Artifact
            The artifact to persist.
        data : bytes
            The raw data to persist as an artifact.
        project_name : str
            The name of the project this artifact belongs to.
        experiment_id : str, optional
            The ID of the experiment this artifact belongs to.
            Artifacts do not need to belong to an experiment.
        """
        artifact_metadata_path = self._get_artifact_metadata_path(
            project_name, experiment_id, artifact.id
        )
        artifact_data_path = self._get_artifact_data_path(project_name, experiment_id, artifact.id)

        self._persist_bytes(data, artifact_data_path)
        self._persist_domain(artifact, artifact_metadata_path)

    def get_artifact_metadata(self, project_name, artifact_id, experiment_id=None):
        """Retrieve an artifact's metadata from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the artifact with ID
            `artifact_id` is logged to.
        artifact_id : str
            The ID of the artifact to retrieve.
        experiment_id : str, optional
            The ID of the experiment the artifact with ID
            `artifact_id` is logged to. Artifacts do not
            need to belong to an experiment.

        Returns
        -------
        rubicon.domain.Artifact
            The artifact with ID `artifact_id`.
        """
        artifact_metadata_path = self._get_artifact_metadata_path(
            project_name, experiment_id, artifact_id
        )

        try:
            open_file = self.filesystem.open(artifact_metadata_path)
        except FileNotFoundError:
            raise RubiconException(f"No artifact with id `{artifact_id}` found.")

        with open_file as f:
            artifact = json.load(f)

        return domain.Artifact(**artifact)

    def get_artifacts_metadata(self, project_name, experiment_id=None):
        """Retrieve all artifacts' metadata from the configured
        filesystem that belong to the specified object.

        Parameters
        ----------
        project_name : str
            The name of the project to retrieve all artifacts
            from.
        experiment_id : str, optional
            The ID of the experiment to retrieve all artifacts
            from. Artifacts do not need to belong to an
            experiment.

        Returns
        -------
        list of rubicon.domain.Artifact
            The artifacts logged to the specified object.
        """
        artifact_metadata_root = self._get_artifact_metadata_root(project_name, experiment_id)

        try:
            artifact_metadata_paths = self._ls_directories_only(artifact_metadata_root)
            artifacts = [
                domain.Artifact(**json.loads(data))
                for data in self.filesystem.cat(artifact_metadata_paths).values()
            ]
        except FileNotFoundError:
            return []

        return artifacts

    def get_artifact_data(self, project_name, artifact_id, experiment_id=None):
        """Retrieve an artifact's raw data.

        Parameters
        ----------
        project_name : str
            The name of the project the artifact with ID
            `artifact_id` is logged to.
        artifact_id : str
            The ID of the artifact to retrieve data from.
        experiment_id : str, optional
            The ID of the experiment the artifact with ID
            `artifact_id` is logged to. Artifacts do not
            need to belong to an experiment.

        Returns
        -------
        bytes
            The artifact with ID `artifact_id`'s raw data.
        """
        artifact_data_path = self._get_artifact_data_path(project_name, experiment_id, artifact_id)

        try:
            open_file = self.filesystem.open(artifact_data_path, "rb")
        except FileNotFoundError:
            raise RubiconException(f"No data for artifact with id `{artifact_id}` found.")

        return open_file.read()

    def delete_artifact(self, project_name, artifact_id, experiment_id=None):
        """Delete an artifact from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the artifact with ID
            `artifact_id` is logged to.
        artifact_id : str
            The ID of the artifact to delete.
        experiment_id : str, optional
            The ID of the experiment the artifact with ID
            `artifact_id` is logged to. Artifacts do not
            need to belong to an experiment.
        """
        artifact_metadata_root = self._get_artifact_metadata_root(project_name, experiment_id)

        try:
            self.filesystem.rm(f"{artifact_metadata_root}/{artifact_id}", recursive=True)
        except FileNotFoundError:
            raise RubiconException(f"No artifact with id `{artifact_id}` found.")

    # ------- Dataframes -------

    def _get_dataframe_metadata_root(self, project_name, experiment_id=None):
        """Returns the dataframes directory of the project with name
        `project_name` or experiment with ID `experiment_id`.
        """
        if experiment_id is not None:
            experiment_metadata_root = self._get_experiment_metadata_root(project_name)

            return f"{experiment_metadata_root}/{experiment_id}/dataframes"
        else:
            return f"{self.root_dir}/{slugify(project_name)}/dataframes"

    def _get_dataframe_metadata_path(self, project_name, experiment_id, dataframe_id):
        """Returns the path of the dataframe with ID `dataframe_id`'s
        metadata.
        """
        dataframe_metadata_root = self._get_dataframe_metadata_root(project_name, experiment_id)

        return f"{dataframe_metadata_root}/{dataframe_id}/metadata.json"

    def _get_dataframe_data_path(self, project_name, experiment_id, dataframe_id):
        """Returns the path of the dataframe with ID `dataframe_id`'s
        raw data.
        """
        dataframe_metadata_root = self._get_dataframe_metadata_root(project_name, experiment_id)

        return f"{dataframe_metadata_root}/{dataframe_id}/data"

    def _convert_to_dask_dataframe(self, df):
        """Converts `df` to a Dask dataframe if it is not already one.
        """
        if not isinstance(df, dd.DataFrame):
            return dd.from_pandas(df, npartitions=1)

        return df

    def _persist_dataframe(self, df, path):
        """Persists the `dask` dataframe `df` to the configured filesystem.
        """
        df.to_parquet(path, engine="pyarrow")

    def _read_dataframe(self, path):
        """Reads the `dask` dataframe `df` from the configured filesystem.
        """
        return dd.read_parquet(path, engine="pyarrow")

    def create_dataframe(self, dataframe, data, project_name, experiment_id=None):
        """Persist a dataframe to the configured filesystem.

        Parameters
        ----------
        dataframe : rubicon.domain.Dataframe
            The dataframe to persist.
        data : dask.dataframe.DataFrame or pandas.DataFrame
            The raw data to persist as a dataframe. All
            dataframes are persisted as `dask` dataframes.
        project_name : str
            The name of the project this dataframe belongs to.
        experiment_id : str, optional
            The ID of the experiment this dataframe belongs to.
            Dataframes do not need to belong to an experiment.
        """
        data = self._convert_to_dask_dataframe(data)

        dataframe_metadata_path = self._get_dataframe_metadata_path(
            project_name, experiment_id, dataframe.id
        )
        dataframe_data_path = self._get_dataframe_data_path(
            project_name, experiment_id, dataframe.id
        )

        self._persist_dataframe(data, dataframe_data_path)
        self._persist_domain(dataframe, dataframe_metadata_path)

    def get_dataframe_metadata(self, project_name, dataframe_id, experiment_id=None):
        """Retrieve a dataframes's metadata from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the dataframe with ID
            `dataframe_id` is logged to.
        dataframe_id : str
            The ID of the dataframe to retrieve.
        experiment_id : str, optional
            The ID of the experiment the dataframe with ID
            `dataframe_id` is logged to. Dataframes do not
            need to belong to an experiment.

        Returns
        -------
        rubicon.domain.Dataframe
            The dataframe with ID `dataframe_id`.
        """
        dataframe_metadata_path = self._get_dataframe_metadata_path(
            project_name, experiment_id, dataframe_id
        )

        try:
            open_file = self.filesystem.open(dataframe_metadata_path)
        except FileNotFoundError:
            raise RubiconException(f"No dataframe with id `{dataframe_id}` found.")

        with open_file as f:
            dataframe = json.load(f)

        return domain.Dataframe(**dataframe)

    def get_dataframes_metadata(self, project_name, experiment_id=None):
        """Retrieve all dataframes' metadata from the configured
        filesystem that belong to the specified object.

        Parameters
        ----------
        project_name : str
            The name of the project to retrieve all dataframes
            from.
        experiment_id : str, optional
            The ID of the experiment to retrieve all dataframes
            from. Dataframes do not need to belong to an
            experiment.

        Returns
        -------
        list of rubicon.domain.Dataframe
            The dataframes logged to the specified object.
        """
        dataframe_metadata_root = self._get_dataframe_metadata_root(project_name, experiment_id)

        try:
            dataframe_metadata_paths = self._ls_directories_only(dataframe_metadata_root)
            dataframes = [
                domain.Dataframe(**json.loads(data))
                for data in self.filesystem.cat(dataframe_metadata_paths).values()
            ]
        except FileNotFoundError:
            return []

        return dataframes

    def get_dataframe_data(self, project_name, dataframe_id, experiment_id=None):
        """Retrieve a dataframe's raw data.

        Parameters
        ----------
        project_name : str
            The name of the project the dataframe with ID
            `dataframe_id` is logged to.
        dataframe_id : str
            The ID of the dataframe to retrieve data from.
        experiment_id : str, optional
            The ID of the experiment the dataframe with ID
            `artifact_id` is logged to. Dataframes do not
            need to belong to an experiment.

        Returns
        -------
        dask.dataframe.DataFrame
            The dataframe with ID `dataframe_id`'s raw data.
        """
        dataframe_data_path = self._get_dataframe_data_path(
            project_name, experiment_id, dataframe_id
        )

        try:
            df = self._read_dataframe(dataframe_data_path)
        except FileNotFoundError:
            raise RubiconException(f"No data for dataframe with id `{dataframe_id}` found.")

        return df

    def delete_dataframe(self, project_name, dataframe_id, experiment_id=None):
        """Delete a dataframe from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the dataframe with ID
            `dataframe_id` is logged to.
        dataframe_id : str
            The ID of the dataframe to delete.
        experiment_id : str, optional
            The ID of the experiment the dataframe with ID
            `artifact_id` is logged to. Dataframes do not
            need to belong to an experiment.
        """
        dataframe_metadata_root = self._get_dataframe_metadata_root(project_name, experiment_id)

        try:
            self.filesystem.rm(f"{dataframe_metadata_root}/{dataframe_id}", recursive=True)
        except FileNotFoundError:
            raise RubiconException(f"No dataframe with id `{dataframe_id}` found.")

    # -------- Features --------

    def _get_feature_metadata_root(self, project_name, experiment_id):
        """Returns the features directory of the experiment with
        ID `experiment_id`.
        """
        experiment_metadata_root = self._get_experiment_metadata_root(project_name)

        return f"{experiment_metadata_root}/{experiment_id}/features"

    def _get_feature_metadata_path(self, project_name, experiment_id, feature_name):
        """Returns the path of the feature with name `feature_name`'s
        metadata.
        """
        feature_metadata_root = self._get_feature_metadata_root(project_name, experiment_id)

        return f"{feature_metadata_root}/{slugify(feature_name)}/metadata.json"

    def create_feature(self, feature, project_name, experiment_id):
        """Persist a feature to the configured filesystem.

        Parameters
        ----------
        feature : rubicon.domain.Feature
            The feature to persist.
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment this feature belongs to.
        """
        feature_metadata_path = self._get_feature_metadata_path(
            project_name, experiment_id, feature.name
        )

        if self.filesystem.exists(feature_metadata_path):
            raise RubiconException(f"A feature with name '{feature.name}' already exists.")

        self._persist_domain(feature, feature_metadata_path)

    def get_feature(self, project_name, experiment_id, feature_name):
        """Retrieve a feature from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment the feature with name
            `feature_name` is logged to.
        feature_name : str
            The name of the feature to retrieve.

        Returns
        -------
        rubicon.domain.Feature
            The feature with name `feature_name`.
        """
        feature_metadata_path = self._get_feature_metadata_path(
            project_name, experiment_id, feature_name
        )

        try:
            open_file = self.filesystem.open(feature_metadata_path)
        except FileNotFoundError:
            raise RubiconException(f"No feature with name '{feature_name}' found.")

        with open_file as f:
            feature = json.load(f)

        return domain.Feature(**feature)

    def get_features(self, project_name, experiment_id):
        """Retrieve all features from the configured filesystem
        that belong to the experiment with ID `experiment_id`.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment to retrieve all features
            from.

        Returns
        -------
        list of rubicon.domain.Feature
            The features logged to the experiment with ID
            `experiment_id`.
        """
        feature_metadata_root = self._get_feature_metadata_root(project_name, experiment_id)

        try:
            feature_metadata_paths = self._ls_directories_only(feature_metadata_root)
            features = [
                domain.Feature(**json.loads(data))
                for data in self.filesystem.cat(feature_metadata_paths).values()
            ]
        except FileNotFoundError:
            return []

        return features

    # -------- Metrics ---------

    def _get_metric_metadata_root(self, project_name, experiment_id):
        """Returns the metrics directory of the experiment with
        ID `experiment_id`.
        """
        experiment_metadata_root = self._get_experiment_metadata_root(project_name)

        return f"{experiment_metadata_root}/{experiment_id}/metrics"

    def _get_metric_metadata_path(self, project_name, experiment_id, metric_name):
        """Returns the path of the metric with name `metric_name`'s
        metadata.
        """
        metric_metadata_root = self._get_metric_metadata_root(project_name, experiment_id)

        return f"{metric_metadata_root}/{slugify(metric_name)}/metadata.json"

    def create_metric(self, metric, project_name, experiment_id):
        """Persist a metric to the configured filesystem.

        Parameters
        ----------
        metric : rubicon.domain.Metric
            The metric to persist.
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment this metric belongs to.
        """
        metric_metadata_path = self._get_metric_metadata_path(
            project_name, experiment_id, metric.name
        )

        if self.filesystem.exists(metric_metadata_path):
            raise RubiconException(f"A metric with name '{metric.name}' already exists.")

        self._persist_domain(metric, metric_metadata_path)

    def get_metric(self, project_name, experiment_id, metric_name):
        """Retrieve a metric from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment the metric with name
            `metric_name` is logged to.
        metric_name : str
            The name of the metric to retrieve.

        Returns
        -------
        rubicon.domain.Metric
            The metric with name `metric_name`.
        """
        metric_metadata_path = self._get_metric_metadata_path(
            project_name, experiment_id, metric_name
        )

        try:
            open_file = self.filesystem.open(metric_metadata_path)
        except FileNotFoundError:
            raise RubiconException(f"No metric with name '{metric_name}' found.")

        with open_file as f:
            metric = json.load(f)

        return domain.Metric(**metric)

    def get_metrics(self, project_name, experiment_id):
        """Retrieve all metrics from the configured filesystem
        that belong to the experiment with ID `experiment_id`.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment to retrieve all metrics
            from.

        Returns
        -------
        list of rubicon.domain.Metric
            The metrics logged to the experiment with ID
            `experiment_id`.
        """
        metric_metadata_root = self._get_metric_metadata_root(project_name, experiment_id)

        try:
            metric_metadata_paths = self._ls_directories_only(metric_metadata_root)
            metrics = [
                domain.Metric(**json.loads(data))
                for data in self.filesystem.cat(metric_metadata_paths).values()
            ]
        except FileNotFoundError:
            return []

        return metrics

    # ------- Parameters -------

    def _get_parameter_metadata_root(self, project_name, experiment_id):
        """Returns the parameters directory of the experiment with
        ID `experiment_id`.
        """
        experiment_metadata_root = self._get_experiment_metadata_root(project_name)

        return f"{experiment_metadata_root}/{experiment_id}/parameters"

    def _get_parameter_metadata_path(self, project_name, experiment_id, parameter_name):
        """Returns the path of the parameter with name `parameter_name`'s
        metadata.
        """
        parameter_metadata_root = self._get_parameter_metadata_root(project_name, experiment_id)

        return f"{parameter_metadata_root}/{slugify(parameter_name)}/metadata.json"

    def create_parameter(self, parameter, project_name, experiment_id):
        """Persist a parameter to the configured filesystem.

        Parameters
        ----------
        parameter : rubicon.domain.Parameter
            The parameter to persist.
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment this parameter belongs to.
        """
        parameter_metadata_path = self._get_parameter_metadata_path(
            project_name, experiment_id, parameter.name
        )

        if self.filesystem.exists(parameter_metadata_path):
            raise RubiconException(f"A parameter with name '{parameter.name}' already exists.")

        self._persist_domain(parameter, parameter_metadata_path)

    def get_parameter(self, project_name, experiment_id, parameter_name):
        """Retrieve a parameter from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project this parameter belongs to.
        experiment_id : str
            The ID of the experiment the parameter with name
            `parameter_name` is logged to.
        parameter_name : str
            The name of the parameter to retrieve.

        Returns
        -------
        rubicon.domain.Parameter
            The parameter with name `parameter_name`.
        """
        parameter_metadata_path = self._get_parameter_metadata_path(
            project_name, experiment_id, parameter_name
        )

        try:
            open_file = self.filesystem.open(parameter_metadata_path)
        except FileNotFoundError:
            raise RubiconException(f"No parameter with name '{parameter_name}' found.")

        with open_file as f:
            parameter = json.load(f)

        return domain.Parameter(**parameter)

    def get_parameters(self, project_name, experiment_id):
        """Retrieve all parameters from the configured filesystem
        that belong to the experiment with ID `experiment_id`.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment to retrieve all parameters
            from.

        Returns
        -------
        list of rubicon.domain.Parameter
            The parameters logged to the experiment with ID
            `experiment_id`.
        """
        parameter_metadata_root = self._get_parameter_metadata_root(project_name, experiment_id)

        try:
            parameter_metadata_paths = self._ls_directories_only(parameter_metadata_root)
            parameters = [
                domain.Parameter(**json.loads(data))
                for data in self.filesystem.cat(parameter_metadata_paths).values()
            ]
        except FileNotFoundError:
            return []

        return parameters

    # ---------- Tags ----------

    def _get_tag_metadata_root(self, project_name, experiment_id=None, dataframe_id=None):
        if dataframe_id is not None:
            dataframe_metadata_root = self._get_dataframe_metadata_root(project_name, experiment_id)

            return f"{dataframe_metadata_root}/{dataframe_id}"
        elif experiment_id is not None:
            experiment_metadata_root = self._get_experiment_metadata_root(project_name)

            return f"{experiment_metadata_root}/{experiment_id}"
        else:
            raise ValueError("`experiment_id` and `dataframe_id` can not both be `None`.")

    def add_tags(self, project_name, tags, experiment_id=None, dataframe_id=None):
        """Persist tags to the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the object to tag
            belongs to.
        tags : list of str
            The tag values to persist.
        experiment_id : str, optional
            The ID of the experiment to apply the tags
            `tags` to.
        dataframe_id : str, optional
            The ID of the dataframe to apply the tags
            `tags` to.
        """
        tag_metadata_root = self._get_tag_metadata_root(project_name, experiment_id, dataframe_id)
        tag_metadata_path = f"{tag_metadata_root}/tags_{domain.utils.uuid.uuid4()}.json"

        self._persist_domain({"added_tags": tags}, tag_metadata_path)

    def remove_tags(self, project_name, tags, experiment_id=None, dataframe_id=None):
        """Delete tags from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the object to delete
            tags from belongs to.
        tags : list of str
            The tag values to delete.
        experiment_id : str, optional
            The ID of the experiment to delete the tags
            `tags` from.
        dataframe_id : str, optional
            The ID of the dataframe to delete the tags
            `tags` from.
        """
        tag_metadata_root = self._get_tag_metadata_root(project_name, experiment_id, dataframe_id)
        tag_metadata_path = f"{tag_metadata_root}/tags_{domain.utils.uuid.uuid4()}.json"

        self._persist_domain({"removed_tags": tags}, tag_metadata_path)

    def _sort_tag_paths(self, tag_paths):
        """Sorts the paths in `tags_paths` by when they were
        created.
        """
        if isinstance(tag_paths, dict):
            tag_paths = tag_paths.values()

        tag_paths_with_timestamps = [
            (t.get("created", t.get("LastModified")), t.get("name")) for t in tag_paths
        ]
        tag_paths_with_timestamps.sort()

        return tag_paths_with_timestamps

    def get_tags(self, project_name, experiment_id=None, dataframe_id=None):
        """Retrieve tags from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the object to retrieve
            tags from belongs to.
        experiment_id : str, optional
            The ID of the experiment to retrieve tags from.
        dataframe_id : str, optional
            The ID of the dataframe to retrieve tags from.

        Returns
        -------
        list of dict
            A list of dictionaries with one key each,
            `added_tags` or `removed_tags`, where the
            value is a list of tag names that have been
            added to or removed from the specified object.
        """
        tag_metadata_root = self._get_tag_metadata_root(project_name, experiment_id, dataframe_id)
        tag_metadata_glob = f"{tag_metadata_root}/tags_*.json"

        tag_paths = self.filesystem.glob(tag_metadata_glob, detail=True)
        if len(tag_paths) == 0:
            return []

        sorted_tag_paths = self._sort_tag_paths(tag_paths)

        tag_data = self.filesystem.cat([p for _, p in sorted_tag_paths])
        sorted_tag_data = [json.loads(tag_data[p]) for _, p in sorted_tag_paths]

        return sorted_tag_data



import json
import uuid

import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
import dash_table
import plotly.graph_objects as go
from dash import callback_context
from dash.dependencies import ALL, MATCH, Input, Output, State

from rubicon.ui.app import app

# INDIVIDUAL PROJECT EXPLORER VIEW


def _get_experiment_table(id, experiments_df):
    """Get a Dash DataTable with the experiments in `experiments_df`.
    """
    return dash_table.DataTable(
        id={"type": "experiment-table", "index": id},
        columns=[{"name": i, "id": i, "selectable": True} for i in experiments_df.columns],
        data=experiments_df.compute().to_dict("records"),
        page_size=10,
        filter_action="native",
        sort_action="native",
        sort_mode="multi",
        row_selectable="multi",
        selected_rows=[],
        fixed_columns={"headers": True, "data": 1},
        style_cell={"overflow": "hidden", "textOverflow": "ellipsis"},
        style_header={"color": "#707171", "fontWeight": 700},
        style_table={"minWidth": "100%"},
        style_cell_conditional=[
            {
                "if": {"column_id": "id"},
                "width": "300px",
                "minWidth": "300px",
                "maxWidth": "300px",
            },
            {
                "if": {"column_id": "commit_hash"},
                "width": "120px",
                "minWidth": "120px",
                "maxWidth": "120px",
            },
        ],
        style_data_conditional=[{"if": {"row_index": "odd"}, "backgroundColor": "#f1f7fa"}],
    )


def _get_github_commit_url(github_url, commit_hash):
    """Get the github commit url, if it exists"""
    is_github_enabled = github_url is not None
    github_commit_url = f"{github_url[:-4]}/tree/{commit_hash}" if is_github_enabled else None

    return github_commit_url


def make_individual_project_explorer_layout(rubicon_model, commit_hash):
    """The html layout for an individual project explorer view determined by
    `commit_hash` in the dashboard.

    A project explorer view shows a subset of experiments logged
    to a project in a tabular format, as well as plotted on a
    parallel coordinates plot.
    """
    id = str(uuid.uuid4())

    experiment_table_df = rubicon_model.get_experiment_table_df(commit_hash)
    github_commit_url = _get_github_commit_url(
        rubicon_model.selected_project.github_url, commit_hash
    )

    group_store = dcc.Store(
        id={"type": "group-store", "index": id}, data={"commit_hash": commit_hash},
    )

    group_preview_title = [
        html.P(
            f"{len(experiment_table_df)} experiments",
            id="group-preview-title",
            className="group-preview-title",
        )
    ]

    if commit_hash is not None and rubicon_model.selected_project.github_url is not None:
        group_preview_title.append(
            html.A(
                f"at commit {commit_hash[:7]}",
                id="group-preview-title-link",
                className="group-preview-title-link",
                href=github_commit_url,
                target="_blank",
            )
        )

    group_model_names = rubicon_model.get_model_names(commit_hash)
    if len(group_model_names) > 0:
        group_model_names_text = f"model name: {group_model_names[0]}"
        if len(group_model_names) > 1:
            group_model_names_text += f" (+{len(group_model_names) - 1} more)"

        group_model_names_view = html.P(
            group_model_names_text,
            id="group-preview-model-names",
            className="group-preview-model-names",
        )
    else:
        group_model_names_view = html.P(style={"display": "none"})

    chevron = html.I(className="fas fa-chevron-down")

    group_preview_row = dbc.Row(
        id={"type": "group-preview-row", "index": id},
        className="group-preview-row",
        children=[
            dbc.Row(group_preview_title, style={"margin": "inherit"}),
            group_model_names_view,
            html.Button(
                chevron,
                id={"type": "show-group-detail-collapsable-btn", "index": id},
                className="show-group-detail-collapsable-btn",
            ),
            html.Button(
                chevron,
                id={"type": "hide-group-detail-collapsable-btn", "index": id},
                className="hide-group-detail-collapsable-btn",
                hidden=True,
            ),
        ],
    )

    experiment_table_bulk_action_button_group = html.Div(
        className="btn-group",
        children=[
            html.Button(
                "Select All",
                id={"type": "select-all-btn", "index": id},
                className="btn-progressive",
            ),
            html.Button(
                "Clear All", id={"type": "clear-all-btn", "index": id}, className="btn-progressive",
            ),
        ],
    )

    group_detail_card = dbc.Card(
        id={"type": "group-detail-card", "index": id},
        className="group-detail-card",
        children=[
            dbc.CardBody(
                id={"type": "group-detail-card-body", "index": id},
                className="group-detail-card-body",
                children=[
                    experiment_table_bulk_action_button_group,
                    _get_experiment_table(id, experiment_table_df),
                    _get_comparison_layout(id, rubicon_model, commit_hash),
                ],
            )
        ],
    )

    group_detail_collapsable = dbc.Collapse(
        id={"type": "group-detail-collapsable", "index": id},
        className="group-detail-collapsable",
        children=[group_detail_card],
    )

    return dbc.Row(
        id={"type": "individual-project-explorer", "index": id},
        className="individual-project-explorer",
        children=[dbc.Col([group_store, group_preview_row, group_detail_collapsable])],
    )


def _get_comparison_layout(id, rubicon_model, commit_hash):
    experiment_comparison_plot = html.Div(
        id={"type": "experiment-comparison-plot", "index": id},
        className="experiment-comparison-plot",
    )

    experiment_comparison_tooltip_icon = html.I(
        id="experiment-comparison-question-tooltip", className="far fa-question-circle"
    )

    experiment_comparison_tooltip_text = dbc.Tooltip(
        """Select experiments within the experiment table and then choose an anchor """
        """to see how the other parameters and metrics affect the anchor's value across experiment runs. """
        """Rearrange the axes by dragging the column names. Select a subset by """
        """clicking and dragging along the y axes.""",
        target="experiment-comparison-question-tooltip",
        style={"max-width": "480px"},
        placement="right",
    )

    experiment_comparison_header = dbc.Row(
        id={"type": "experiment-comparison-header", "index": id},
        className="experiment-comparison-header",
        children=[
            "Compare Experiments",
            experiment_comparison_tooltip_icon,
            experiment_comparison_tooltip_text,
        ],
    )

    anchor_options = rubicon_model.get_anchor_options(commit_hash)

    anchor_dropdown = dcc.Dropdown(
        id={"type": "anchor-dropdown", "index": id},
        className="anchor-dropdown",
        placeholder="Select an anchor metric",
        clearable=False,
        options=anchor_options,
        value=anchor_options[-1]["value"],
    )

    return html.Div([experiment_comparison_header, anchor_dropdown, experiment_comparison_plot])


# GROUPED PROJECT EXPLORER VIEW


@app.callback(
    [
        Output({"type": "group-detail-collapsable", "index": MATCH}, "is_open"),
        Output({"type": "show-group-detail-collapsable-btn", "index": MATCH}, "hidden"),
        Output({"type": "hide-group-detail-collapsable-btn", "index": MATCH}, "hidden"),
    ],
    [
        Input({"type": "show-group-detail-collapsable-btn", "index": MATCH}, "n_clicks_timestamp"),
        Input({"type": "hide-group-detail-collapsable-btn", "index": MATCH}, "n_clicks_timestamp"),
    ],
)
def _toggle_experiment_group_collapsable(last_show_click, last_hide_click):
    """The callback to show and collapse each group.

    Also toggles the show/hide buttons. Triggered when the show or
    hide button is clicked.
    """
    last_show_click = last_show_click if last_show_click else 0
    last_hide_click = last_hide_click if last_hide_click else 0

    # "show" is clicked: open collabsable, hide "show" button & un-hide "hide" button
    if int(last_show_click) > int(last_hide_click):
        return True, True, False
    # "hide" is clicked: close collabsable, un-hide "show" button & hide "hide" button
    elif int(last_hide_click) > int(last_show_click):
        return False, False, True
    # nothing yet is clicked: return default states
    else:
        return False, False, True


@app.callback(
    Output({"type": "experiment-comparison-plot", "index": MATCH}, "children"),
    [
        Input({"type": "experiment-table", "index": MATCH}, "derived_virtual_data"),
        Input({"type": "experiment-table", "index": MATCH}, "derived_virtual_selected_rows"),
        Input({"type": "anchor-dropdown", "index": MATCH}, "value"),
    ],
    [State({"type": "group-store", "index": MATCH}, "data")],
)
def _update_experiment_comparison_plot(
    experiment_table_data, experiment_table_selected_rows, anchor, data
):
    """The callback to render a new experiment comparison plot.

    Triggered when new rows in the experiment table are selected.
    or deselected.
    """
    if experiment_table_selected_rows is None or len(experiment_table_selected_rows) == 0:
        return [html.Div()]

    commit_hash = data["commit_hash"]
    selected_experiment_ids = [
        experiment_table_data[row]["id"] for row in experiment_table_selected_rows
    ]

    anchor_data, dimensions = app._rubicon_model.get_dimensions(
        commit_hash, selected_experiment_ids, anchor
    )

    return [
        dcc.Graph(
            figure=go.Figure(
                go.Parcoords(
                    line=dict(color=anchor_data, colorscale="plasma", showscale=True),
                    dimensions=dimensions,
                )
            )
        )
    ]


@app.callback(
    Output("grouped-project-explorer", "children"),
    [Input({"type": "project-selection--project", "index": ALL}, "n_clicks")],
)
def _update_project_explorer(values):
    """The callback to render the grouped project explorer.

    Triggered when a project is selected in the project selection list.
    """
    # if all values are 0, the user hasn't clicked a project yet
    is_waiting_for_first_click = True
    for value in values:
        if value != 0:
            is_waiting_for_first_click = False

    if is_waiting_for_first_click:
        return make_empty_view("Please select a project to view.")

    # use `dash.callback_context` to get the id of the clicked project list item
    selected_id = callback_context.triggered[0]["prop_id"].split(".")[0]
    selected_project_name = json.loads(selected_id)["index"]

    app._rubicon_model.update_selected_project(selected_project_name)

    project_explorer_header = dbc.Row(
        id="experiment-deatils-header",
        className="experiment-details-header",
        children=selected_project_name,
    )

    experiment_groups = app._rubicon_model._experiment_table_dfs.items()

    # handle no experiments view
    if len(experiment_groups) == 0:
        return [project_explorer_header, make_empty_view("Log some experiments to this project!")]

    _project_explorers = [
        make_individual_project_explorer_layout(app._rubicon_model, group)
        for group, _ in experiment_groups
    ]

    return [project_explorer_header, *_project_explorers]


@app.callback(
    Output({"type": "experiment-table", "index": MATCH}, "selected_rows"),
    [
        Input({"type": "select-all-btn", "index": MATCH}, "n_clicks_timestamp"),
        Input({"type": "clear-all-btn", "index": MATCH}, "n_clicks_timestamp"),
    ],
    [State({"type": "experiment-table", "index": MATCH}, "derived_virtual_indices")],
)
def _update_selected_experiment_table_rows(
    last_select_click, last_clear_click, experiment_table_indices
):
    """The callback to select or deselect all rows in the experiment table.

    Triggered when the select all or clear all button is clicked.
    """
    last_select_click = last_select_click if last_select_click else 0
    last_clear_click = last_clear_click if last_clear_click else 0

    # "select all" is clicked: return all row indicies
    if int(last_select_click) > int(last_clear_click):
        return experiment_table_indices

    # "clear all" or nothing yet is clicked: return no row indicies
    return []


def make_empty_view(instructions):
    """The html layout for the dashboard's empty view when no project is selected.
    """
    title = "Nothing to see here..."

    return html.Div(
        id="empty-view",
        className="empty-view",
        children=[
            html.Img(src=app.get_asset_url("search-icon.svg"), style={"marginBottom": "20px"}),
            html.P(className="empty-view--title", children=title),
            html.P(className="empty-view--instructions", children=instructions),
        ],
    )


def make_project_explorer_layout():
    """The html layout for the dashboard's grouped project explorer view.

    This view holds a collection of collapsable project explorer
    groups. Each group shows some top level information and
    provides interactions to drill into the underlying group's data.
    """
    return dbc.Col(id="grouped-project-explorer", className="grouped-project-explorer")



from rubicon.repository.base import BaseRepository

from rubicon.repository.local import LocalRepository
from rubicon.repository.memory import MemoryRepository
from rubicon.repository.s3 import S3Repository

__all__ = [
    "BaseRepository",
    "LocalRepository",
    "MemoryRepository",
    "S3Repository",
]



import subprocess

import fsspec
import yaml

from rubicon import domain
from rubicon.client import Config, Project
from rubicon.exceptions import RubiconException
from rubicon.repository.utils import slugify


class Rubicon:
    """The `rubicon` client's entry point.

    Creates a `Config` and injects it into the client level
    objects at run-time.

    Parameters
    ----------
    persistence : str, optional
        The persistence type. Can be one of ["filesystem", "memory"].
        Defaults to "filesystem".
    root_dir : str, optional
        Absolute or relative filepath. Use absolute path for best performance.
        Defaults to the local filesystem. Prefix with s3:// to use s3 instead.
    auto_git_enabled : bool, optional
        True to use the `git` command to automatically log relevant repository
        information to projects and experiments logged with this client instance,
        False otherwise. Defaults to False.
    """

    def __init__(self, persistence="filesystem", root_dir=None, auto_git_enabled=False):
        self.config = Config(persistence, root_dir, auto_git_enabled)

    @property
    def repository(self):
        return self.config.repository

    @repository.setter
    def repository(self, value):
        self.config.repository = value

    def _get_github_url(self):
        """Returns the repository URL of the `git` repo it is called from.
        """
        completed_process = subprocess.run(["git", "remote", "-v"], capture_output=True)
        remotes = completed_process.stdout.decode("utf8").replace("\t", " ").split("\n")

        try:
            origin = [remote for remote in remotes if remote.startswith("origin")][0]
            github_url = origin.split(" ")[1]
        except IndexError:
            github_url = None

        return github_url

    def _create_project_domain(self, name, description, github_url, training_metadata):
        """Instantiates and returns a project domain object.
        """
        if self.config.is_auto_git_enabled and github_url is None:
            github_url = self._get_github_url()

        if training_metadata is not None:
            training_metadata = domain.utils.TrainingMetadata(training_metadata)

        return domain.Project(
            name,
            description=description,
            github_url=github_url,
            training_metadata=training_metadata,
        )

    def create_project(self, name, description=None, github_url=None, training_metadata=None):
        """Create a project.

        Parameters
        ----------
        name : str
            The project's name.
        description : str, optional
            The project's description.
        github_url : str, optional
            The URL of the GitHub repository associated with this
            project. If omitted and automatic `git` logging is
            enabled, it will be retrieved via `git remote`.
        training_metadata : tuple or list of tuples, optional
            Metadata associated with the training dataset(s)
            used across each experiment in this project.

        Returns
        -------
        rubicon.client.Project
            The created project.
        """
        project = self._create_project_domain(name, description, github_url, training_metadata)
        self.repository.create_project(project)

        return Project(project, self.config)

    def get_project(self, name):
        """Get a project.

        Parameters
        ----------
        name : str
            The name of the project to get.

        Returns
        -------
        rubicon.client.Project
            The project with name `name`.
        """
        project = self.repository.get_project(name)

        return Project(project, self.config)

    def get_project_as_dask_df(self, name, group_by=None):
        """Get a dask dataframe representation of a project.

        Parameters
        ----------
        name : str
            The name of the project to get.
        group_by : str or None, optional
            How to group the project's experiments in the returned
            DataFrame(s). Valid options include ["commit_hash"].

        Returns
        -------
        dask.DataFrame or list of dask.DataFrame
            If `group_by` is `None`, a dask dataframe holding the project's
            data. Otherwise a list of dask dataframes holding the project's
            data grouped by `group_by`.
        """
        project = self.get_project(name)

        return project.to_dask_df(group_by=None)

    def get_or_create_project(self, name, **kwargs):
        """Get or create a project.

        Parameters
        ----------
        name : str
            The project's name.
        kwargs : dict
            Additional keyword arguments to be passed to
            `Rubicon.create_project`.

        Returns
        -------
        rubicon.client.Project
            The corresponding project.
        """
        try:
            project = self.get_project(name)
        except RubiconException:
            project = self.create_project(name, **kwargs)

        return project

    def projects(self):
        """Get a list of available projects.

        Returns
        -------
        list of rubicon.client.Project
            The list of available projects.
        """
        return [Project(project, self.config) for project in self.repository.get_projects()]

    def publish(
        self,
        project_name,
        experiment_ids=None,
        experiment_tags=None,
        experiment_qtype="or",
        output_filepath=None,
    ):
        """Publish a project and its experiments to an intake
        catalog that can be read by the `intake-rubicon` driver.

        Parameters
        ----------
        project_name : str
            The name of the project to publish.
        experiment_ids : list of str, optional
            The ids of the experiments belonging to the project
            named `project_name` to publish.
        experiment_tags : list of str, optional
            The tags of the experiments belonging to the project
            named `project_name` to publish.
        experiment_qtype : str, optional
            The query type to filter experiment tags on. Can be
            'or' or 'and'. Defaults to 'or'.
        output_filepath : str, optional
            The absolute or relative local filepath or S3
            bucket and key to log the output catalog YAML file
            to. S3 buckets must be prepended with 's3://'. If
            left as None, only returns the YAML string without
            writing.

        Notes
        -----
        This function is not threadsafe.

        If neither `experiment_ids` nor `experiment_tags` are
        provided, all experiments belonging to the project
        named `project_name` will be published. If both are
        provided, only `experiment_ids` will be used.

        Returns
        -------
        str
            The YAML string representation of the catalog.
        """
        catalog = {"sources": {}}

        project = self.get_project(project_name)
        project_catalog = {
            "driver": "rubicon_project",
            "args": {"urlpath": self.repository.root_dir, "project_name": project.name},
        }

        catalog["sources"][f"project_{project.id.replace('-', '_')}"] = project_catalog

        if experiment_ids is not None:
            experiments = []
            for e_id in experiment_ids:
                experiments.append(project.experiment(e_id))
        elif experiment_tags is not None:
            experiments = project.experiments(tags=experiment_tags, qtype=experiment_qtype)
        else:
            experiments = project.experiments()

        for e in experiments:
            experiment_catalog = {
                "driver": "rubicon_experiment",
                "args": {
                    "urlpath": self.repository.root_dir,
                    "project_name": project.name,
                    "experiment_id": e.id,
                },
            }

            catalog["sources"][f"experiment_{e.id.replace('-', '_')}"] = experiment_catalog

        catalog_yaml = yaml.dump(catalog)

        if output_filepath is not None:
            with fsspec.open(output_filepath, "w", auto_mkdir=False) as f:
                f.write(catalog_yaml)

        return catalog_yaml

    def sync(self, project_name, s3_root_dir):
        """Sync a local project to S3.

        Parameters
        ----------
        project_name : str
            The name of the project to sync.
        s3_root_dir : str
            The S3 path where the project's data
            will be synced.

        Notes
        -----
        Use to backup your local project data to S3, as an alternative to direct S3 logging.
        Relies on AWS CLI's sync. Ensure that your credentials are set and that your Proxy
        is on.
        """
        if self.config.persistence != "filesystem":
            raise RubiconException(
                "You can't sync projects written to memory. Sync from either local filesystem or S3."
            )

        project = self.get_project(project_name)
        local_path = f"{self.config.root_dir}/{slugify(project.name)}"
        cmd = f"aws s3 sync {local_path} {s3_root_dir}/{slugify(project.name)}"

        try:
            subprocess.run(cmd, shell=True, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            raise RubiconException(e.stderr)



import dataclasses
import json
from datetime import datetime

from rubicon.domain.utils import TrainingMetadata


class DomainJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        """
        Note - if we need to support nested objects
        within dataclasses, we need to leverage asdict()
        """
        if isinstance(obj, datetime):
            return {"_type": "datetime", "value": str(obj)}
        if isinstance(obj, set):
            return {"_type": "set", "value": list(obj)}
        if isinstance(obj, TrainingMetadata):
            return {"_type": "training_metadata", "value": obj.training_metadata}
        if dataclasses.is_dataclass(obj):
            return obj.__dict__
        else:
            return super().default(obj)  # pragma: no cover


class DomainJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if obj.get("_type") == "datetime":
            return datetime.strptime(obj.get("value"), "%Y-%m-%d %H:%M:%S.%f")
        if obj.get("_type") == "set":
            return set(obj.get("value"))
        if obj.get("_type") == "training_metadata":
            return TrainingMetadata([(*o,) for o in obj.get("value")])
        else:
            return obj


def dumps(data):
    return json.dumps(data, cls=DomainJSONEncoder)


def load(open_file):
    return json.load(open_file, cls=DomainJSONDecoder)


def loads(data):
    return json.loads(data, cls=DomainJSONDecoder)



import os
import subprocess

from rubicon.exceptions import RubiconException
from rubicon.repository import LocalRepository, MemoryRepository, S3Repository


class Config:
    """Used to configure `rubicon` client objects.

    Configuration can be specified (in order of precedence) by:
        1. environment variables 'PERSISTENCE' and 'ROOT_DIR'
        2. arguments to `__init__`

    Parameters
    ----------
    persistence : str, optional
        The persistence type. Can be one of ["filesystem", "memory"].
    root_dir : str, optional
        Absolute or relative filepath. Defaults to using the local
        filesystem. Prefix with s3:// to use s3 instead.
    auto_git_enabled : bool, optional
        True to use the `git` command to automatically log relevant repository
        information to projects and experiments logged with this client instance,
        False otherwise. Defaults to False.
    """

    PERSISTENCE_TYPES = ["filesystem", "memory"]
    REPOSITORIES = {
        "memory-memory": MemoryRepository,
        "filesystem-local": LocalRepository,
        "filesystem-s3": S3Repository,
    }

    def __init__(self, persistence=None, root_dir=None, is_auto_git_enabled=False):
        self.persistence, self.root_dir, self.is_auto_git_enabled = self._load_config(
            persistence, root_dir, is_auto_git_enabled
        )

        self.repository = self._get_repository()

    def _check_is_in_git_repo(self):
        """Raise a `RubiconException` if not called from within a `git` repository.
        """
        if subprocess.run(["git", "rev-parse", "--git-dir"], capture_output=True).returncode != 0:
            raise RubiconException(
                "Not a `git` repo: Falied to locate the '.git' directory in this or any parent directories."
            )

    def _load_config(self, persistence, root_dir, is_auto_git_enabled):
        """Get the configuration values.
        """
        persistence = os.environ.get("PERSISTENCE", persistence)
        if persistence not in self.PERSISTENCE_TYPES:
            raise ValueError(f"PERSISTENCE must be one of {self.PERSISTENCE_TYPES}.")

        root_dir = os.environ.get("ROOT_DIR", root_dir)
        if root_dir is None and persistence != "memory":
            raise ValueError("root_dir cannot be None.")

        if is_auto_git_enabled:
            self._check_is_in_git_repo()

        return (persistence, root_dir, is_auto_git_enabled)

    def _get_protocol(self):
        """Get the file protocol of the configured root directory.
        """
        if self.persistence == "memory":
            return "memory"
        elif self.persistence == "filesystem":
            if self.root_dir.startswith("s3://"):
                return "s3"
            else:
                return "local"

    def _get_repository(self):
        """Get the repository for the configured persistence type.
        """
        protocol = self._get_protocol()

        repository_key = f"{self.persistence}-{protocol}"
        repository = self.REPOSITORIES.get(repository_key)

        if repository is None:
            raise RubiconException(
                f"{self.__class__.__module__}.{self.__class__.__name__} has no persistence "
                + f"layer for the provided configuration: `persistence`: {self.persistence}, "
                + f"`protocol` (from `root_dir`): {protocol}"
            )

        return repository(self.root_dir)



from rubicon import domain
from rubicon.client import (
    ArtifactMixin,
    Base,
    DataframeMixin,
    Feature,
    Metric,
    Parameter,
    TagMixin,
)


class Experiment(Base, ArtifactMixin, DataframeMixin, TagMixin):
    """A client experiment.

    An `experiment` represents a model run and is identified by
    its 'created_at' time. It can have `metrics`, `parameters`,
    `features`, `dataframes`, and `artifacts` logged to it.

    An `experiment` is logged to a `project`.

    Parameters
    ----------
    domain : rubicon.domain.Experiment
        The experiment domain model.
    parent : rubicon.client.Project
        The project that the experiment is logged to.
    """

    def __init__(self, domain, parent):
        super().__init__(domain, parent._config)

        self._parent = parent
        self._artifacts = []
        self._dataframes = []
        self._metrics = []
        self._features = []
        self._parameters = []

    def log_metric(self, name, value, directionality="score", description=None):
        """Create a metric under the experiment.

        Parameters
        ----------
        name : str
            The metric's name.
        value : float
            The metric's value.
        directionality : str, optional
            The metric's directionality. Must be one of
            ["score", "loss"], where "score" represents
            a metric to maximize, while "loss" represents a
            metric to minimize. Defaults to "score".
        description : str, optional
            The metric's description. Use to provide additional
            context.

        Returns
        -------
        rubicon.client.Metric
            The created metric.
        """
        metric = domain.Metric(name, value, directionality=directionality, description=description)
        self.repository.create_metric(metric, self.project.name, self.id)

        return Metric(metric, self._config)

    def metrics(self):
        """Get the metrics logged to this experiment.

        Returns
        -------
        list of rubicon.client.Metric
            The metrics previously logged to this experiment.
        """
        self._metrics = [
            Metric(m, self._config) for m in self.repository.get_metrics(self.project.name, self.id)
        ]

        return self._metrics

    def log_feature(self, name, description=None, importance=None):
        """Create a feature under the experiment.

        Parameters
        ----------
        name : str
            The features's name.
        description : str
            The feature's description. Use to provide
            additional context.
        importance : float
            The feature's importance.

        Returns
        -------
        rubicon.client.Feature
            The created feature.
        """
        feature = domain.Feature(name, description=description, importance=importance)
        self.repository.create_feature(feature, self.project.name, self.id)

        return Feature(feature, self._config)

    def features(self):
        """Get the features logged to this experiment.

        Returns
        -------
        list of rubicon.client.Feature
            The features previously logged to this experiment.
        """
        self._features = [
            Feature(f, self._config)
            for f in self.repository.get_features(self.project.name, self.id)
        ]

        return self._features

    def log_parameter(self, name, value=None, description=None):
        """Create a parameter under the experiment.

        Parameters
        ----------
        name : str
            The parameter's name.
        value : object, optional
            The parameter's value. Can be an object of any JSON
            serializable (via rubicon.utils.DomainJSONEncoder)
            type.
        description : str, optional
            The parameter's description. Use to provide
            additional context.

        Returns
        -------
        rubicon.client.Parameter
            The created parameter.
        """
        parameter = domain.Parameter(name, value=value, description=description)
        self.repository.create_parameter(parameter, self.project.name, self.id)

        return Parameter(parameter, self._config)

    def parameters(self):
        """Get the parameters logged to this experiment.

        Returns
        -------
        list of rubicon.client.Parameter
            The parameters previously logged to this experiment.
        """
        self._parameters = [
            Parameter(p, self._config)
            for p in self.repository.get_parameters(self.project.name, self.id)
        ]

        return self._parameters

    @property
    def id(self):
        """Get the experiment's id.
        """
        return self._domain.id

    @property
    def name(self):
        """Get the experiment's name.
        """
        return self._domain.name

    @property
    def description(self):
        """Get the experiment's description.
        """
        return self._domain.description

    @property
    def model_name(self):
        """Get the experiment's model name.
        """
        return self._domain.model_name

    @property
    def branch_name(self):
        """Get the experiment's branch name.
        """
        return self._domain.branch_name

    @property
    def commit_hash(self):
        """Get the experiment's commit hash.
        """
        return self._domain.commit_hash

    @property
    def training_metadata(self):
        """Get the project's training metadata.
        """
        training_metadata = self._domain.training_metadata.training_metadata

        if len(training_metadata) == 1:
            training_metadata = training_metadata[0]

        return training_metadata

    @property
    def created_at(self):
        """Get the time the experiment was created.
        """
        return self._domain.created_at

    @property
    def project(self):
        """Get the project client object that this experiment
        belongs to.
        """
        return self._parent



import os
import pickle

import fsspec

from rubicon.repository import LocalRepository


class MemoryRepository(LocalRepository):
    """The memory repository persists Rubicon data to an
    in-memory filesystem representation.

    The memory repository is intended for testing and
    development purposes. Data persisted to the memory
    filesystem will not be persisted between runtimes.

    Parameters
    ----------
    root_dir : str, optional
        A path representing the virtual root directory for
        the in-memory filesystem. This does not need to be
        specified unless interacting with an already created
        in-memory filesystem.
    """

    PROTOCOL = "memory"

    def __init__(self, root_dir=None):
        self.filesystem = fsspec.filesystem(self.PROTOCOL)
        self.root_dir = root_dir.rstrip("/") if root_dir is not None else "/root"

        self.filesystem.mkdir(self.root_dir)

    def _add_dirnames_to_pseudo_dirs(self, path):
        """Adds each persisted in-memory file's directory to
        `fsspec`'s `pseudo_dirs`.

        This is a workaround to a bug in `fsspec` where the
        memory filesystem does not properly populate the
        `pseudo_dirs` list.
        """
        pseudo_dirs = set(self.filesystem.pseudo_dirs)

        while path != "" and path != "/":
            path = os.path.dirname(path)
            pseudo_dirs.add(path)

        self.filesystem.pseudo_dirs = list(pseudo_dirs)

    def _persist_bytes(self, bytes_data, path):
        """Persists the Rubicon object `domain` to the
        in-memory path defined by `path`.
        """
        super()._persist_bytes(bytes_data, path)

        self._add_dirnames_to_pseudo_dirs(path)

    def _persist_domain(self, domain, path):
        """Persists the Rubicon object `domain` to the
        in-memory path defined by `path`.
        """
        super()._persist_domain(domain, path)

        self._add_dirnames_to_pseudo_dirs(path)

    def _persist_dataframe(self, df, path):
        """Persists the `dask` dataframe `df` to the in-memory
        path defined by `path`.
        """
        with self.filesystem.open(path, "wb") as f:
            pickle.dump(df, f)

    def _read_dataframe(self, path):
        """Reads the `dask` dataframe `df` from the in-memory
        path defined by `path`.
        """
        with self.filesystem.open(path, "rb") as f:
            data = pickle.load(f)

        return data



from rubicon.exceptions import RubiconException


class TrainingMetadata:
    """A tuple or list of tuples representing metadata
    about a training dataset. Intentionally left arbitrary
    to allow users to use as it suits them.

    Notes
    -----
    `TrainingMetadata` is a simple wrapper around a
    list of tuples for the purpose of validation and
    de/serialization.

    Examples
    --------
    >>> TrainingMetadata([
    >>>     ("s3://bucket/a.parquet", "SELECT * FROM a"),
    >>>     ("s3://bucket/b.parquet", "SELECT * FROM b"),
    >>> ])
    [('s3://bucket/a.parquet', 'SELECT * FROM a'), ...]
    >>> TrainingMetadata(
    >>>     ("s3", ["bucket/a.csv", "bucket/b.csv"], "SELECT * FROM x")
    >>> )
    [('s3', ['bucket/a.csv', 'bucket/b.csv'], 'SELECT * FROM x')]
    """

    def __init__(self, training_metadata):
        if not isinstance(training_metadata, list):
            training_metadata = [training_metadata]

        if not all([isinstance(tm, tuple) for tm in training_metadata]):
            raise RubiconException("`training_metadata` must be a list of tuples.")

        self.training_metadata = training_metadata

    def __repr__(self):
        return str(self.training_metadata)



import asyncio
from datetime import datetime

from rubicon import domain
from rubicon.client import asynchronous as client
from rubicon.client.mixin import MultiParentMixin
from rubicon.client.utils.tags import has_tag_requirements


class ArtifactMixin(MultiParentMixin):
    """Adds artifact support to an asynchronous client object.
    """

    async def log_artifact(
        self, data_bytes=None, data_file=None, data_path=None, name=None, description=None
    ):
        """Overrides `rubicon.client.ArtifactMixin.log_artifact` to
        asynchronously log an artifact to this client object.

        Parameters
        ----------
        data_bytes : bytes, optional
            The raw bytes to log as an artifact.
        data_file : TextIOWrapper, optional
            The open file to log as an artifact.
        data_path : str, optional
            The absolute or relative local path or S3 path
            to the data to log as an artifact. S3 paths
            must be prepended with 's3://'.
        name : str, optional
            The name of the artifact file. Required if
            `data_path` is not provided.
        description : str, optional
            A description of the artifact. Use to provide
            additional context.

        Notes
        -----
        Only one of `data_bytes`, `data_file`, and `data_path`
        should be provided. If more than one is given, the order
        of precedence is `data_bytes`, `data_file`, `data_path`.

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.

        Examples
        --------
        >>> # Log with bytes
        >>> await experiment.log_artifact(
        ...     data_bytes=b'hello rubicon!', name='bytes_artifact', description="log artifact from bytes"
        ... )

        >>> # Log with file
        >>> with open('some_relevant_file', 'rb') as f:
        >>>     await project.log_artifact(
        ...         data_file=f, name='file_artifact', description="log artifact from file"
        ...     )

        >>> # Log with file path
        >>> await experiment.log_artifact(
        ...     data_path="./path/to/artifact.pkl", description="log artifact from file path"
        ... )
        """
        data_bytes, name = self._validate_data(data_bytes, data_file, data_path, name)

        artifact = domain.Artifact(name=name, description=description, parent_id=self._domain.id)

        project_name, experiment_id = self._get_parent_identifiers()
        await self.repository.create_artifact(
            artifact, data_bytes, project_name, experiment_id=experiment_id
        )

        return client.Artifact(artifact, self)

    async def log_conda_environment(self, artifact_name=None):
        """Log the conda environment as an artifact to this client object.
        Useful for recreating your exact environment at a later date.

        Parameters
        ----------
        artifact_name : str, optional
            The name of the artifact (the exported conda environment).

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.

        Notes
        -----
        Relies on running with an active conda environment.
        """
        if artifact_name is None:
            artifact_name = f"environment-{datetime.now().strftime('%Y_%m_%d-%I_%M_%S_%p')}.yml"

        env_bytes = self._get_environment_bytes("conda env export")
        artifact = await self.log_artifact(data_bytes=env_bytes, name=artifact_name)

        return artifact

    async def log_pip_requirements(self, artifact_name=None):
        """Log the pip requirements as an artifact to this client object.
        Useful for recreating your exact environment at a later date.

        Parameters
        ----------
        artifact_name : str, optional
            The name of the artifact (the exported pip environment).

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.
        """
        if artifact_name is None:
            artifact_name = f"requirements-{datetime.now().strftime('%Y_%m_%d-%I_%M_%S_%p')}.txt"

        requirements_bytes = self._get_environment_bytes("pip freeze")
        artifact = await self.log_artifact(data_bytes=requirements_bytes, name=artifact_name)

        return artifact

    async def artifacts(self):
        """Overrides `rubicon.client.ArtifcatMixin.artifacts` to
        asynchronously get the artifacts logged to this client object.

        Returns
        -------
        list of rubicon.client.Artifact
            The artifacts previously logged to this client object.
        """
        project_name, experiment_id = self._get_parent_identifiers()

        self._artifacts = [
            client.Artifact(a, self)
            for a in await self.repository.get_artifacts_metadata(
                project_name, experiment_id=experiment_id
            )
        ]

        return self._artifacts

    async def delete_artifacts(self, ids):
        """Overrides `rubicon.client.ArtifcatMixin.delete_artifacts` to
        asynchronously delete the artifacts logged to with client
        objects with ids `ids`.

        Parameters
        ----------
        ids : list of str
            The ids of the artifacts to delete.
        """
        project_name, experiment_id = self._get_parent_identifiers()

        await asyncio.gather(
            *[
                self.repository.delete_artifact(
                    project_name, artifact_id, experiment_id=experiment_id
                )
                for artifact_id in ids
            ]
        )


class DataframeMixin(MultiParentMixin):
    """Adds dataframe support to an asynchronous client object.
    """

    async def log_dataframe(self, df, description=None, tags=[]):
        """Overrides `rubicon.client.DataframeMixin.log_dataframe` to
        asynchronously log a dataframe to this client object.

        Parameters
        ----------
        df : pandas.DataFrame or dask.dataframe.DataFrame
            The `dask` or `pandas` dataframe to log.
        description : str, optional
            The dataframe's description. Use to provide
            additional context.
        tags : list of str
            The values to tag the dataframe with.

        Returns
        -------
        rubicon.client.Dataframe
            The new dataframe.
        """
        dataframe = domain.Dataframe(parent_id=self._domain.id, description=description, tags=tags)

        project_name, experiment_id = self._get_parent_identifiers()
        await self.repository.create_dataframe(
            dataframe, df, project_name, experiment_id=experiment_id
        )

        return client.Dataframe(dataframe, self)

    async def _filter_dataframes(self, dataframes, tags, qtype):
        """Filters the provided dataframes by `tags` using
        query type `qtype`.
        """
        if len(tags) > 0:
            filtered_dataframes = []
            [
                filtered_dataframes.append(d)
                for d in dataframes
                if has_tag_requirements(await d.tags, tags, qtype)
            ]
            self._dataframes = filtered_dataframes
        else:
            self._dataframes = dataframes

    async def dataframes(self, tags=[], qtype="or"):
        """Overrides `rubicon.client.DataframeMixin.dataframes` to
        asynchronously get the dataframes logged to this client object.

        Parameters
        ----------
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Dataframe
            The dataframes previously logged to this client object.
        """
        project_name, experiment_id = self._get_parent_identifiers()
        dataframes = [
            client.Dataframe(d, self)
            for d in await self.repository.get_dataframes_metadata(
                project_name, experiment_id=experiment_id
            )
        ]

        await self._filter_dataframes(dataframes, tags, qtype)

        return self._dataframes

    async def delete_dataframes(self, ids):
        """Overrides `rubicon.client.DataframeMixin.delete_dataframe`
        to asynchronously delete the dataframes with ids `ids` logged
        to this client object.

        Parameters
        ----------
        ids : list of str
            The ids of the dataframes to delete.
        """
        project_name, experiment_id = self._get_parent_identifiers()

        await asyncio.gather(
            *[
                self.repository.delete_dataframe(
                    project_name, dataframe_id, experiment_id=experiment_id
                )
                for dataframe_id in ids
            ]
        )


class TagMixin:
    """Adds tag support to an asynchronous client object.
    """

    async def add_tags(self, tags):
        """Overrides `rubicon.client.TagMixin.add_tags` to
        asynchronously add tags to this client object.

        Parameters
        ----------
        tags : list of str
            The tag values to add.
        """
        project_name, experiment_id, dataframe_id = self._get_taggable_identifiers()

        self._domain.add_tags(tags)
        await self.repository.add_tags(
            project_name, tags, experiment_id=experiment_id, dataframe_id=dataframe_id
        )

    async def remove_tags(self, tags):
        """Overrides `rubicon.client.TagMixin.remove_tags` to
        asynchronously remove tags from this client object.

        Parameters
        ----------
        tags : list of str
             The tag values to remove.
        """
        project_name, experiment_id, dataframe_id = self._get_taggable_identifiers()

        self._domain.remove_tags(tags)
        await self.repository.remove_tags(
            project_name, tags, experiment_id=experiment_id, dataframe_id=dataframe_id
        )

    @property
    async def tags(self):
        """Overrides `rubicon.client.TagMixin.tags` to
        asynchronously get this client object's tags.
        """
        project_name, experiment_id, dataframe_id = self._get_taggable_identifiers()
        tag_data = await self.repository.get_tags(
            project_name, experiment_id=experiment_id, dataframe_id=dataframe_id,
        )

        self._update_tags(tag_data)

        return self._domain.tags



from rubicon.client import Parameter as SyncParameter


class Parameter(SyncParameter):
    """An asycnhronous client parameter.

    A `parameter` is an input to an `experiment` (model run)
    that depends on the type of model being used. It affects
    the model's predictions.

    For example, if you were using a random forest classifier,
    'n_estimators' (the number of trees in the forest) could
    be a parameter.

    A `parameter` is logged to an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Parameter
        The parameter domain model.
    config : rubicon.client.Config
        The config, which specifies the underlying repository.
    """

    pass



import dash_html_components as html


def make_footer_layout():
    """The html layout for the dashboard's footer view.
    """
    return html.Div(
        id="footer",
        className="footer",
        children=[
            html.Div(id="copyright", className="footer--copyright", children="© 2020 Capital One"),
        ],
    )



from rubicon.client.asynchronous.config import Config

from rubicon.client.asynchronous.mixin import ArtifactMixin, DataframeMixin, TagMixin  # noqa F401

from rubicon.client.asynchronous.artifact import Artifact
from rubicon.client.asynchronous.dataframe import Dataframe
from rubicon.client.asynchronous.feature import Feature
from rubicon.client.asynchronous.metric import Metric
from rubicon.client.asynchronous.parameter import Parameter
from rubicon.client.asynchronous.experiment import Experiment
from rubicon.client.asynchronous.project import Project
from rubicon.client.asynchronous.rubicon import Rubicon

__all__ = [
    "Artifact",
    "Config",
    "Dataframe",
    "Experiment",
    "Feature",
    "Metric",
    "Parameter",
    "Project",
    "Rubicon",
]



from prefect import task
from rubicon import Rubicon


@task
def get_or_create_project_task(
    persistence, root_dir, project_name, auto_git_enabled=False, **kwargs
):
    """Get or create a project within a `prefect` flow.

    This `prefect` task can be used within a flow to
    create a new project or get an existing one. It should
    be the entry point to any `prefect` flow that logs
    data to Rubicon.

    Parameters
    ----------
    persistence : str
        The persistence type to be passed to the
        `Rubicon` constructor.
    root_dir : str
        The root directory to be passed to the
        `Rubicon` constructor.
    project_name : str
        The name of the project to get or create.
    auto_git_enabled : bool, optional
        True to use the `git` command to automatically log
        relevant repository information to projects and
        experiments logged with the client instance created
        in this task, False otherwise. Defaults to False.
    kwargs : dict
        Additional keyword arguments to be passed to
        `Rubicon.create_project`.

    Returns
    -------
    rubicon.client.Project
        The project with name `project_name`.
    """
    rubicon = Rubicon(persistence=persistence, root_dir=root_dir, auto_git_enabled=auto_git_enabled)
    project = rubicon.get_or_create_project(project_name, **kwargs)

    return project


@task
def create_experiment_task(project, **kwargs):
    """Create an experiment within a `prefect` flow.

    This `prefect` task can be used within a flow to
    create a new experiment under an existing project.

    Parameters
    ----------
    project : rubicon.client.Project
        The project under which the experiment will be created.
    kwargs : dict
        Keyword arguments to be passed to
        `Project.log_experiment`.

    Returns
    -------
    rubicon.client.Experiment
        The created experiment.
    """
    return project.log_experiment(**kwargs)


@task
def log_artifact_task(parent, **kwargs):
    """Log an artifact within a `prefect` flow.

    This `prefect` task can be used within a flow to
    log an artifact to an existing project or experiment.

    Parameters
    ----------
    parent : rubicon.client.Project or rubicon.client.Experiment
        The project or experiment to log the artifact to.
    kwargs : dict
        Keyword arguments to be passed to
        `Project.log_artifact` or `Experiment.log_artifact`.

    Returns
    -------
    rubicon.client.Artifact
        The logged artifact.
    """
    return parent.log_artifact(**kwargs)


@task
def log_dataframe_task(parent, df, **kwargs):
    """Log a dataframe within a `prefect` flow.

    This `prefect` task can be used within a flow to
    log a dataframe to an existing project or experiment.

    Parameters
    ----------
    parent : rubicon.client.Project or rubicon.client.Experiment
        The project or experiment to log the dataframe to.
    df : pandas.DataFrame or dask.dataframe.DataFrame
        The `pandas` or `dask` dataframe to log.
    kwargs : dict
        Additional keyword arguments to be passed to
        `Project.log_dataframe` or `Experiment.log_dataframe`.

    Returns
    -------
    rubicon.client.Dataframe
        The logged dataframe.
    """
    return parent.log_dataframe(df, **kwargs)


@task
def log_feature_task(experiment, feature_name, **kwargs):
    """Log a feature within a `prefect` flow.

    This `prefect` task can be used within a flow to
    log a feature to an existing experiment.

    Parameters
    ----------
    experiment : rubicon.client.Experiment
        The experiment to log a new feature to.
    feature_name : str
        The name of the feature to log. Passed to
        `Experiment.log_feature` as `name`.
    kwargs : dict
        Additional keyword arguments to be passed to
        `Experiment.log_feature`.

    Returns
    -------
    rubicon.client.Feature
        The logged feature.
    """
    return experiment.log_feature(feature_name, **kwargs)


@task
def log_metric_task(experiment, metric_name, metric_value, **kwargs):
    """Log a metric within a `prefect` flow.

    This `prefect` task can be used within a flow to
    log a metric to an existing experiment.

    Parameters
    ----------
    experiment : rubicon.client.Experiment
        The experiment to log a new metric to.
    metric_name : str
        The name of the metric to log. Passed to
        `Experiment.log_metric` as `name`.
    metric_value : str
        The value of the metric to log. Passed to
        `Experiment.log_metric` as `value`.
    kwargs : dict
        Additional keyword arguments to be passed to
        `Experiment.log_metric`.

    Returns
    -------
    rubicon.client.Metric
        The logged metric.
    """
    return experiment.log_metric(metric_name, metric_value, **kwargs)


@task
def log_parameter_task(experiment, parameter_name, parameter_value, **kwargs):
    """Log a parameter within a `prefect` flow.

    This `prefect` task can be used within a flow to
    log a parameter to an existing experiment.

    Parameters
    ----------
    experiment : rubicon.client.Experiment
        The experiment to log a new parameter to.
    parameter_name : str
        The name of the parameter to log. Passed to
        `Experiment.log_parameter` as `name`.
    parameter_value : str
        The value of the parameter to log. Passed to
        `Experiment.log_parameter` as `value`.
    kwargs : dict
        Additional keyword arguments to be passed to
        `Experiment.log_parameter`.

    Returns
    -------
    rubicon.client.Parameter
        The logged parameter.
    """
    return experiment.log_parameter(parameter_name, parameter_value, **kwargs)



import click

from rubicon.ui import Dashboard


@click.group()
@click.version_option()
def cli():
    pass


# Top level CLI commands


@cli.command()
@click.option(
    "--root-dir",
    type=click.STRING,
    help="The absolute path to the top level folder holding the Rubicon project(s).",
    required=True,
)
@click.option(
    "--host", "-h", type=click.STRING, help="The host to serve the dashboard on.", default=None
)
@click.option(
    "--port", "-p", type=click.STRING, help="The port to serve the dashboard on.", default=None
)
@click.option(
    "--debug", "-d", type=click.BOOL, help="Whether or not to run in debug mode.", default=False
)
def ui(root_dir, host, port, debug):
    """Launch the Rubicon Dashboard.
    """
    dashboard = Dashboard("filesystem", root_dir)

    server_kwargs = dict(debug=debug, port=port, host=host)
    server_kwargs = {k: v for k, v in server_kwargs.items() if v is not None}
    dashboard.run_server(**server_kwargs)


# CLI groups

if __name__ == "__main__":
    cli()



from rubicon import __version__
from rubicon.client import Rubicon
from rubicon.intake_rubicon.base import DataSourceMixin


class ProjectSource(DataSourceMixin):
    """An Intake data source for reading `rubicon` projects.

    Parameters
    ----------
    urlpath : str
        The root directory the `rubicon` data is logged to.
    project_name : str
        The name of the `rubicon` project to load.
    """

    version = __version__

    container = "python"
    name = "rubicon_project"

    def _get_schema(self):
        """Load the project named `self._project_name`.
        """
        self._rubicon = Rubicon(persistence="filesystem", root_dir=self._urlpath)
        self._rubicon_object = self._rubicon.get_project(self._project_name)

        self._metadata.update(
            {
                "project": {
                    "name": self._rubicon_object.name,
                    "id": self._rubicon_object.id,
                    "description": self._rubicon_object.description,
                    "created_id": self._rubicon_object.created_at,
                }
            }
        )

        return super()._get_schema()



from rubicon.client import Base


class Feature(Base):
    """A client feature.

    A `feature` is an input to an `experiment` (model run)
    that's an independent, measurable property of a phenomenon
    being observed. It affects the model's predictions.

    For example, consider a model that predicts how likely a
    customer is to pay back a loan. Possible features could be
    'age', 'credit score', etc.

    A `feature` is logged to an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Feature
        The feature domain model.
    config : rubicon.client.Config
        The config, which specifies the underlying repository.
    """

    def __init__(self, domain, config=None):
        super().__init__(domain, config)

    @property
    def id(self):
        """Get the feature's id.
        """
        return self._domain.id

    @property
    def name(self):
        """Get the feature's name.
        """
        return self._domain.name

    @property
    def description(self):
        """Get the feature's description.
        """
        return self._domain.description

    @property
    def importance(self):
        """Get the feature's importance.
        """
        return self._domain.importance

    @property
    def created_at(self):
        """Get the feature's created_at.
        """
        return self._domain.created_at



from rubicon.repository.asynchronous.base import AsynchronousBaseRepository
from rubicon.repository.asynchronous.s3 import S3Repository

__all__ = [
    "AsynchronousBaseRepository",
    "S3Repository",
]



import dash_bootstrap_components as dbc
import dash_html_components as html

from rubicon.ui.app import app
from rubicon.ui.model import RubiconModel
from rubicon.ui.views.footer import make_footer_layout
from rubicon.ui.views.header import make_header_layout
from rubicon.ui.views.project_explorer import make_project_explorer_layout
from rubicon.ui.views.project_selection import make_project_selection_layout


class Dashboard:
    """The Rubicon Dashboard.

    Parameters
    ----------
    persistence : str
        The persistence type. Can be one of ["filesystem", "memory"].
    root_dir : str, optional
        Absolute or relative filepath of the root directory holding Rubicon data.
        Use absolute path for best performance. Defaults to the local filesystem.
        Prefix with s3:// to use s3 instead.
    """

    def __init__(self, persistence, root_dir=None):
        self._app = app
        self.rubicon_model = RubiconModel(persistence, root_dir)
        self._app._rubicon_model = self.rubicon_model

        self._app.layout = html.Div(
            [
                dbc.Row(make_header_layout()),
                dbc.Row(
                    [
                        dbc.Col(make_project_selection_layout(), width=2),
                        dbc.Col(
                            make_project_explorer_layout(),
                            width=10,
                            style={
                                "overflowY": "scroll",
                                "maxHeight": "90vh",
                                "paddingRight": "165px",
                            },
                        ),
                    ],
                ),
                dbc.Row(make_footer_layout()),
            ]
        )

    def run_server(self, **kwargs):
        """Serve the dash app.

        Parameters
        ----------
        kwargs : dict
            Arguments passed to dash.run_server()
        """
        self._app.run_server(**kwargs)



from __future__ import annotations

from rubicon.domain.artifact import Artifact
from rubicon.domain.dataframe import Dataframe
from rubicon.domain.experiment import Experiment
from rubicon.domain.feature import Feature
from rubicon.domain.metric import Metric
from rubicon.domain.parameter import Parameter
from rubicon.domain.project import Project

__all__ = ["Artifact", "Dataframe", "Experiment", "Feature", "Metric", "Parameter", "Project"]



from rubicon.client import Feature as SyncFeature


class Feature(SyncFeature):
    """An asynchronous client feature.

    A `feature` is an input to an `experiment` (model run)
    that's an independent, measurable property of a phenomenon
    being observed. It affects the model's predictions.

    For example, consider a model that predicts how likely a
    customer is to pay back a loan. Possible features could be
    'age', 'credit score', etc.

    A `feature` is logged to an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Feature
        The feature domain model.
    config : rubicon.client.Config
        The config, which specifies the underlying repository.
    """

    pass



from setuptools import find_packages, setup

install_requires = [
    "click>=7.1",
    "dask[dataframe]>=2.12.0",
    "fsspec>=0.8.3",
    "intake>=0.5.2",
    "pyarrow>=0.16.0,<0.18.0",
    "pyyaml>=3.12.0,<5.4.0",
    "s3fs>=0.5.1",
]

prefect_requires = ["prefect>=0.12.0"]
ui_requires = ["dash>=1.14.0", "dash-bootstrap-components>=0.10.6"]

all_requires = prefect_requires + ui_requires

extras_require = {
    "all": all_requires,
    "prefect": prefect_requires,
    "ui": ui_requires,
}

setup(
    name="rubicon-ml",
    version="0.1.1",
    author="Joe Wolfe, Ryan Soley, Diane Lee, Mike McCarty, CapitalOne",
    license='Apache License, Version 2.0',
    description="an ML library for model development and governance",
    packages=find_packages(),
    include_package_data=True,
    url="https://github.com/capitalone/rubicon",
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        "console_scripts": ["rubicon=rubicon.cli:cli"],
        "intake.drivers": [
            "rubicon_project = rubicon.intake_rubicon.project:ProjectSource",
            "rubicon_experiment = rubicon.intake_rubicon.experiment:ExperimentSource",
        ],
    },
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',

        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Documentation',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 3 or both.
        'Programming Language :: Python :: 3',
    ],
)



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime

from rubicon.domain.utils import TrainingMetadata, uuid


@dataclass
class Project:
    name: str

    id: str = field(default_factory=uuid.uuid4)
    description: str = None
    github_url: str = None
    training_metadata: TrainingMetadata = None
    created_at: datetime = field(default_factory=datetime.utcnow)



from rubicon.client import Metric as SyncMetric


class Metric(SyncMetric):
    """An asynchronous client metric.

    A `metric` is a single-value output of an `experiment` that
    helps evaluate the quality of the model's predictions.

    It can be either a 'score' (value to maximize) or
    a 'loss' (value to minimize).

    A `metric` is logged to an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Metric
        The metric domain model.
    config : rubicon.client.Config
        The config, which specifies the underlying repository.
    """

    pass
